/*
 * Domination v1.0 by Sami Kyostila, March 1999
 */


#include <stdio.h>
#include <ncurses.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <ctype.h>
#include <signal.h>

#include "vt.h"
#include "ipc_msg.h"
#include "game.h"
#include "arg.h"

#define USERFILE "./.domination_data"
#define MAX_USERS 256

#define PLAYER_1 42
#define PLAYER_2 43
#define M_SERVER 1
#define M_CLIENT 2


typedef struct {
	char name[16];
	int win, lose, quit;
} user_data;


void usage();


int  msgqueue_id, mode, state;

void signal_break(int a)
{
	fprintf(stderr, "Break, aborting...\n\r");
	ipc_remove_queue(msgqueue_id);
	exit(1);
}

void update_userdata(char *user, int win, int lose, int quit)
{
	FILE *handle = fopen(USERFILE, "r+");
		
	int i, found=0;
	long offset=0;

	user_data u;
	memset(&u, 0, sizeof(user_data));

        if (handle == NULL)
        {
                fprintf(stderr, "Unable to update data file.\n\r");
                return;
        }

	while (!feof(handle))
	{
		offset = ftell(handle);
		fread(&u, sizeof(user_data), 1, handle);
		if (strcmp(u.name, user)==0)
		{
			found=1;
			break;
		}
	}
	if (!found)
	{
		fseek(handle, 0, SEEK_END);
		offset = ftell(handle);
		memset(&u, 0, sizeof(user_data));
		u.win = 0;
		u.lose = 0;
		u.quit = 0;
	}

	fseek(handle, offset, SEEK_SET);
	strcpy(u.name, user);
	u.win += win;
	u.lose += lose;
	u.quit += quit;
	fwrite(&u, sizeof(user_data), 1, handle);	
	
	fclose(handle);
}

void view_top()
{
        FILE *handle = fopen(USERFILE, "r");

        int i=0, j, k;
        long offset=0, high=0;

        user_data *u = (user_data*)malloc(MAX_USERS*sizeof(user_data));
        memset(u, 0, sizeof(user_data)*MAX_USERS);

        if (handle == NULL)
        {
                fprintf(stderr, "Unable to open data file.\n\r");
                return;
        }

	
	for (;;)
        {
                fread(&u[i], sizeof(user_data), 1, handle);
		if (feof(handle)) break;
		if (u[i].win > high) high=u[i].win;
		i++;
	}
	printf("%d users.\n\r", i);
	k = 1;

	for (;high>=0; high--)
	for (j=0; j<i; j++)
		if (u[j].win==high)
		{
			if (k==1) printf(VT_BOLD);
			printf("%3d. %16s: won %4d, lost %4d, quit %4d\n\r", k++, u[j].name, u[j].win, u[j].lose, u[j].quit);
			if (k==2) printf(VT_NORMAL);
			if (k>20) break;
		}

	free(u);
	fclose(handle);
}



int main(int argc, char *argv[])
{
	int i, r_tx, r_ty, r_sx, r_sy, initial_pieces=5;
        key_t key;
	char buffer[80];
	char local_mark;

	char local_name[80], remote_name[80];
	cuserid(local_name);

	srand(time(NULL));

	if (arg_exists(argc,argv, "-t"))
	{
		view_top();
		exit(1);
	}

	if (!arg_exists(argc,argv, "-s") && !arg_exists(argc,argv, "-c"))
        	usage();

	if (arg_exists(argc,argv, "-k"))
		keyset = arg_string(argc,argv, "-k");
	
	printf("Using keyset %s.\n\r", keyset);

	if (arg_exists(argc,argv, "-s")) key=atoi(arg_string(argc,argv, "-s"));
	if (arg_exists(argc,argv, "-c")) key=atoi(arg_string(argc,argv, "-c"));

	if (arg_exists(argc,argv, "-p")) initial_pieces=atoi(arg_string(argc,argv, "-p"));

	printf("Using key %ld.\n\r", key);

        if((msgqueue_id = ipc_init(key)) == -1)
	{
                perror("msgget");
                exit(1);
        }

	if (arg_exists(argc,argv, "-s"))
	{
		printf("%s%s%s is playing as server.\n", VT_BOLD, local_name, VT_NORMAL);
		mode = M_SERVER;
		fflush(stdout);
	} else
	if (arg_exists(argc,argv, "-c"))
	{
		printf("%s%s%s is playing as client.\n", VT_BOLD, local_name, VT_NORMAL);
		mode = M_CLIENT;
		fflush(stdout);
	}

	if (mode == M_SERVER)
	{
		printf("\nTelling client about ourselves...\n");
		ipc_send_message(msgqueue_id, PLAYER_1, local_name);

		printf("Waiting for client's name...\n");
                ipc_read_message(msgqueue_id, PLAYER_2);
                strcpy(remote_name, ipc_qbuf.mtext);
                printf("Playing against %s%s%s\n", VT_BOLD, remote_name, VT_NORMAL);

		printf("Generating and broadcasting board with %d pieces...\n", initial_pieces);
		grid_random_init(initial_pieces);
		ipc_send_message(msgqueue_id, PLAYER_1, (char*)&grid[0][0]);
	} else {
		printf("\nWaiting for server's name...\n");
		ipc_read_message(msgqueue_id, PLAYER_1);
		strcpy(remote_name, ipc_qbuf.mtext);
		printf("Playing against %s%s%s\n", VT_BOLD, remote_name, VT_NORMAL);

 		printf("Telling server about ourselves...\n");
		ipc_send_message(msgqueue_id, PLAYER_2, local_name);

		printf("Receiving board layout...\n");
		ipc_read_message(msgqueue_id, PLAYER_1);
		strcpy((char*)&grid[0][0], ipc_qbuf.mtext);
	}

	printf("\nGame Initialized.");
	signal(SIGINT, signal_break);


        for (i=0; i<25; i++) printf("\r\n");
	fflush(stdout);

	initscr(); cbreak(); noecho(); nonl();
	refresh();

	vt_cursor_set(50, 19);
	printf("Domination by Sami Kystil\r\n");
	vt_cursor_set(50, 20);
	printf("%c/%c/%c/%c/space/q\r\n", keyset[0], keyset[1], keyset[2], keyset[3]);
	
	vt_cursor_set(1, 20);

	if (mode==M_SERVER) local_mark=PLAYER1; else local_mark=PLAYER2;

	update_state(); print_status(local_name, local_mark, remote_name);

	state = 0;
	if (mode==M_SERVER)
		grid_render(PLAYER1);
	else
		grid_render(PLAYER2);

	fflush(stdout);
	fflush(stdin);

	for (;;)
	{
		if (mode == M_SERVER)
		{
			grid_render(PLAYER1);
			state = place_mark(local_mark);
			update_state();
			grid_render(PLAYER1);
			print_status(local_name, local_mark, remote_name);
			if (!state)
			{
				sprintf(buffer, "%c%c%c%c", (char)mark_x+1, (char)mark_y+1, (char)sx+1, (char)sy+1);
				ipc_send_message(msgqueue_id, PLAYER_1, buffer);
			} else {
				buffer[0] = 0xff;
				ipc_send_message(msgqueue_id, PLAYER_1, buffer);
				status("\x7User abort!");        
				update_userdata(local_name, 0, 0, 1);
                                break;
			}

			if (check_for_win() == PLAYER1)
			{
				status("\x7You won!");
				update_userdata(local_name, 1, 0, 0);
				break;
			}
			if (check_for_win() == BLANKCHAR)
                        {
				status("\x7It's a tie game!");
                                break;
                        }
			status("Opponent's turn");
			vt_cursor_set(mark_x*2+1, mark_y+1);
                        fflush(stdout);
                        ipc_read_message(msgqueue_id, PLAYER_2);
                        r_tx = ipc_qbuf.mtext[0]-1;
                        r_ty = ipc_qbuf.mtext[1]-1;
                        r_sx = ipc_qbuf.mtext[2]-1;
                        r_sy = ipc_qbuf.mtext[3]-1;

                        if (r_tx==-2)
                        {
				status("\x7Remote user abort!");
				break;
                        }

			process_move(r_sx, r_sy, r_tx, r_ty, PLAYER2); 
			update_state();
			print_status(local_name, local_mark, remote_name);

                        if (check_for_win() == PLAYER2)
                        {
				grid_render(PLAYER1);
				status("\x7You lost!");
				update_userdata(local_name, 0, 1, 0);
                                break;
                        }
			if (check_for_win() == BLANKCHAR)
                        {
				grid_render(PLAYER1);
				status("\x7It's a tie game!");
                                break;
                        }
		} else {
			status("Opponent's turn");
			vt_cursor_set(mark_x*2+1, mark_y+1);
			fflush(stdout);
                	ipc_read_message(msgqueue_id, PLAYER_1);
                        r_tx = ipc_qbuf.mtext[0]-1;
                        r_ty = ipc_qbuf.mtext[1]-1;
                        r_sx = ipc_qbuf.mtext[2]-1;
                        r_sy = ipc_qbuf.mtext[3]-1;

			if (r_tx==-2)
			{
				status("Remote user abort!");
                                break;
			}

			process_move(r_sx, r_sy, r_tx, r_ty, PLAYER1); 
			update_state();
			print_status(local_name, local_mark, remote_name);

                        if (check_for_win() == PLAYER1)
                        {
				grid_render(PLAYER2);
				status("\x7You lost!");
				update_userdata(local_name, 0, 1, 0);
                                break;
                        }
			if (check_for_win() == BLANKCHAR)
                        {
				grid_render(PLAYER2);
				status("\x7It's a tie game!");
                                break;
                        }

			grid_render(PLAYER2);
                        state = place_mark(PLAYER2);
			update_state();
			grid_render(PLAYER2);
			print_status(local_name, local_mark, remote_name);

			if (!state)
                        {
				sprintf(buffer, "%c%c%c%c", (char)mark_x+1, (char)mark_y+1, (char)sx+1, (char)sy+1);
                                ipc_send_message(msgqueue_id, PLAYER_2, buffer);
                        } else {
                                buffer[0] = 0xff;
                                ipc_send_message(msgqueue_id, PLAYER_2, buffer);		
				status("User abort!");
				update_userdata(local_name, 0, 0, 1);
                                break;
			}

			if (check_for_win() == PLAYER2)
                        {
				status("\x7You won!");
				update_userdata(local_name, 1, 0, 0);
                                break;
                        }
			if (check_for_win() == BLANKCHAR)
                        {
				status("\x7It's a tie game!");
                                break;
                        }
		}
		if (state) break;
	}
	nocbreak(); echo();

	vt_cursor_set(1,24);
	printf("Terminating connection...");
	fflush(stdout);

	sleep(1);
        ipc_remove_queue(msgqueue_id);

	printf("done.\n\r\n\r");

	endwin();
	return(0);
}

void usage(void)
{
        fprintf(stderr, "\n%sdom%s - A game of Domination by Sami Kyostila, %s\n", VT_BOLD, VT_NORMAL, __DATE__);
        fprintf(stderr, "%susage:%s dom [-s <key>] | [-c <key>] [-t] [-p <pieces>] [-k <keyset>]\n", VT_BOLD, VT_NORMAL);

	fprintf(stderr, "\n"
			"-s <key>          - Makes this terminal the server\n"
			"-c <key>          - Makes this terminal the client\n"
			"-t                - Lists the top-20 players\n"
			"-p <pieces>       - Set the default number of pieces (server only)\n"
			"-k <keyset>       - Choose an alternate keyset. (optional)\n"
			"                    Each character specifies a single key, in the order of\n"
			"                    left, right, up, down. (e.g '-k jlik')\n"
			);


	fprintf(stderr,	"%sGame rules:\n%s"
			"  The object of the game is to wipe out your opponent's pieces off the board,\n"
			"or to have more pieces than your opponent on the board when the last vacant\n"
			"space is filled up.\n"
			"  During a turn the player must move one of his/her pieces 1 or 2 grid units\n"
			"in any compass direction (N,NE,E,SE,S,SW,W,NW). If he/she decides to move 1\n"
			"unit, the piece is not moved, but rather duplicated at the target grid square.\n"
			"Moving 2 units causes the piece to actually be moved. When a move is made, all\n"
			"the opponent's pieces surrounding the target grid square are transformed in to\n"
			"pieces of the person who made the move.\n"
			"  To move a piece, move the cursor over it and press space, and then move the\n"
			"cursor to the target square (possible targets marked with '+') and press space.\n\n"
			,VT_BOLD,VT_NORMAL);
       exit(1);
}

