#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <string.h>

#include "game.h"

char *keyset = "adws";
int mark_x=GRIDXSIZE/2, mark_y=GRIDYSIZE/2;
int sx,sy;
int p1_count=0, p2_count=0, t_count=0;

void grid_render(char mark)
{
	int x,y,f=0;
	char line[512], tmp[10];

        printf("\x1B[0m\x1B[1;1H");

	for (y=0; y<GRIDYSIZE; y++)
	{
		memset(line, 0, 511);
		for (x=0; x<GRIDXSIZE; x++)
		{
			if (grid[x][y]==mark)
			{
				if (f==0) strcat(line, VT_BOLD);
				f=1;
			}
			if (grid[x][y]!=mark)
			{
				if (f==1) strcat(line, VT_NORMAL);
				f=0;
			}
			tmp[0]=grid[x][y];
			tmp[1]=' ';
			tmp[2]='\0';
			strcat(line, tmp);
		}
		printf("%s\n\r", line);
	}
	printf(VT_NORMAL);
	fflush(stdout);
}


void draw_text(int x, int y, char *text)
{
	char buffer[256];
	sprintf(buffer, "\x1B[%d;%dH%s", y+1, x+1, text);
	printf(buffer);
/*	fflush(stdout);*/
}


void print_status(char *local_name, char mark, char *remote_name)
{
	vt_cursor_set(1,19);

	if (mark==PLAYER1)
	{
		printf("[%s%c%s] %s%16s%s: %4d pieces (%.2f%%)   \r\n", 
			VT_BOLD, PLAYER1, VT_NORMAL, VT_BOLD, local_name, VT_NORMAL, p1_count, (float)(p1_count)/(float)(t_count)*100.0);
		printf("[%c] %16s: %4d pieces (%.2f%%)   \r\n", 
			PLAYER2, remote_name, p2_count, (float)(p2_count)/(float)(t_count)*100.0);
	}
	else
	{
		printf("[%c] %16s: %4d pieces (%.2f%%)   \r\n", 
			PLAYER1, remote_name, p1_count, (float)(p1_count)/(float)(t_count)*100.0);
		printf("[%s%c%s] %s%16s%s: %4d pieces (%.2f%%)   \r\n", 
			VT_BOLD, PLAYER2, VT_NORMAL, VT_BOLD, local_name, VT_NORMAL, p2_count, (float)(p2_count)/(float)(t_count)*100.0);
	}
}

void grid_draw_mark(int x, int y)
{
	char text[8];

	printf("\x1B[0m");
	text[0] = grid[x][y];
	text[1] = '\0';

	draw_text(x*2, y, text);
	printf("\x1B[1D");
/*	fflush(stdout);*/
}

void grid_draw_inverse_mark(int x, int y)
{
	char text[8];

	printf(VT_REVERSE);
	text[0] = grid[x][y];
	text[1] = '\0';

	draw_text(x*2, y, text);
	printf(VT_NORMAL);
/*	fflush(stdout);*/
}

void grid_draw_bold_mark(int x, int y)
{
	char text[8];

	printf(VT_BOLD);
	text[0] = grid[x][y];
	text[1] = '\0';

	draw_text(x*2, y, text);
	printf(VT_NORMAL);
/*	fflush(stdout);*/
}

void grid_draw_blinking_mark(int x, int y)
{
	char text[8];

	printf(VT_BLINK);
	text[0] = grid[x][y];
	text[1] = '\0';

	draw_text(x*2, y, text);
	printf(VT_NORMAL);
/*	fflush(stdout);*/
}


char check_for_mark(int x, int y, char mark)
{
	if (x<0) return 0;
	if (y<0) return 0;
	if (x>GRIDXSIZE-1) return 0;
	if (y>GRIDYSIZE-1) return 0;

	if (grid[x][y]==mark) return 1;
	return 0;
}


char check_for_win()
{
	int x,y,flag=1;

	update_state();
	if (p1_count==0) return PLAYER2;
	if (p2_count==0) return PLAYER1;	

	for (y=0;y<GRIDYSIZE;y++)
	for (x=0;x<GRIDXSIZE;x++)
		if (grid[x][y]==BLANKCHAR) return '\0';

	if (p1_count==p2_count) return BLANKCHAR;
	return (p1_count>p2_count)?PLAYER1:PLAYER2;
}


void grid_init()
{
	memset(grid, BLANKCHAR, GRIDXSIZE*GRIDYSIZE);
}

void grid_random_init(int pieces)
{
	int i=pieces, x,y;
	
	grid_init();

	while(i--)
	{
		x=rand()%(GRIDXSIZE/2);
		y=rand()%GRIDYSIZE;
		while (grid[GRIDXSIZE/2-x][y]!=BLANKCHAR || 
		       grid[GRIDXSIZE/2+x][y]!=BLANKCHAR ||
			x==0)
		{
			x=rand()%(GRIDXSIZE/2);
			y=rand()%GRIDYSIZE;
		}
		grid[GRIDXSIZE/2-x][y]=PLAYER1;
		grid[GRIDXSIZE/2+x][y]=PLAYER2;
	}
	update_state();
}

void status(char *msg)
{
	int i;
	vt_cursor_set(1,23);
	for (i=0; i<75; i++) printf(" ");
	vt_cursor_set(1,23);
	printf(msg);
	fflush(stdout);
}

void update_state(void)
{
	int x,y;
	p1_count=p2_count=t_count=0;

	for (y=0; y<GRIDYSIZE; y++)
	for (x=0; x<GRIDXSIZE; x++)
	{
		if (grid[x][y]!=BLANKCHAR) t_count++;
		if (grid[x][y]==PLAYER1) p1_count++;
		if (grid[x][y]==PLAYER2) p2_count++;
	}
}

void process_move(int sx, int sy, int tx, int ty, char mark)
{
	int x1=tx-1, y1=ty-1, x2=tx+1, y2=ty+1, x,y;
	grid[tx][ty]=mark;

	if (x1<0) x1=0;
	if (y1<0) y1=0;
	if (x2>GRIDXSIZE-1) x2=GRIDXSIZE-1;
	if (y2>GRIDYSIZE-1) y2=GRIDYSIZE-1;

	if (abs(tx-sx)==2 || abs(ty-sy)==2)
		grid[sx][sy]=BLANKCHAR;

	if (grid[x1][y1]!=mark && grid[x1][y1]!=BLANKCHAR) grid[x1][y1]=mark;
	if (grid[x2][y1]!=mark && grid[x2][y1]!=BLANKCHAR) grid[x2][y1]=mark;
	if (grid[x1][y2]!=mark && grid[x1][y2]!=BLANKCHAR) grid[x1][y2]=mark;
	if (grid[x2][y2]!=mark && grid[x2][y2]!=BLANKCHAR) grid[x2][y2]=mark;

	if (grid[tx][y1]!=mark && grid[tx][y1]!=BLANKCHAR) grid[tx][y1]=mark;
	if (grid[tx][y2]!=mark && grid[tx][y2]!=BLANKCHAR) grid[tx][y2]=mark;
	if (grid[x1][ty]!=mark && grid[x1][ty]!=BLANKCHAR) grid[x1][ty]=mark;
	if (grid[x2][ty]!=mark && grid[x2][ty]!=BLANKCHAR) grid[x2][ty]=mark;

	update_state();

	if (!player_can_move(PLAYER1) && p1_count>0)
		for(y=0;y<GRIDYSIZE;y++)
		for(x=0;x<GRIDYSIZE;x++)
			if (grid[x][y]==BLANKCHAR) grid[x][y]=PLAYER2;

	if (!player_can_move(PLAYER2) && p2_count>0)
		for(y=0;y<GRIDYSIZE;y++)
		for(x=0;x<GRIDYSIZE;x++)
			if (grid[x][y]==BLANKCHAR) grid[x][y]=PLAYER1;

	update_state();
}

int is_possible_move(int mx, int my, int px, int py)
{
	if (mx==px && my==py) return 0;
	if (mx<px-2 || my<py-2 || mx>px+2 || my>py+2) return 0;
	if (grid[mx][my]==BLANKCHAR) return 1;
	return 0;
}

void show_possible_moves(int mx, int my)
{
	int x1=mx-2, y1=my-2, x2=mx+2, y2=my+2, x,y;
	if (x1<0) x1=0;
	if (y1<0) y1=0;
	if (x2>GRIDXSIZE-1) x2=GRIDXSIZE-1;
	if (y2>GRIDYSIZE-1) y2=GRIDYSIZE-1;

	for (x=x1;x<=x2; x++)
	for (y=y1;y<=y2; y++)
		if (is_possible_move(x,y, mx,my))
		{
			vt_cursor_set(x*2+1,y+1);
			printf("+");
		}
}

int count_possible_moves(int mx, int my)
{
	int x1=mx-2, y1=my-2, x2=mx+2, y2=my+2, x,y,c=0;

	if (x1<0) x1=0;
	if (y1<0) y1=0;
	if (x2>GRIDXSIZE-1) x2=GRIDXSIZE-1;
	if (y2>GRIDYSIZE-1) y2=GRIDYSIZE-1;

	for (x=x1;x<=x2; x++)
	for (y=y1;y<=y2; y++)
		if (is_possible_move(x,y, mx,my)) c++;

	return c;
}

int player_can_move(char mark)
{
	int mx,my;

	for (my=0; my<GRIDYSIZE; my++)
	for (mx=0; mx<GRIDXSIZE; mx++)
		if (count_possible_moves(mx,my)&&grid[mx][my]==mark) return 1;

	return 0;
}

int place_mark(char mark)
{
	char key;
	int selected=0,x,y;

	status("Select piece to move.");

	grid_draw_inverse_mark(mark_x, mark_y);
	vt_cursor_set((mark_x*2+1), (mark_y+1));
	fflush(stdout);

	for (;;)
	{
		key = getch();

		if (grid[mark_x][mark_y]!=mark)
			grid_draw_mark(mark_x, mark_y);
		else
			grid_draw_bold_mark(mark_x, mark_y);

		if (selected)
		{
			grid_draw_inverse_mark(sx,sy);
			show_possible_moves(sx,sy);
		}		

		if (key == keyset[0]) mark_x--;
		if (key == keyset[1]) mark_x++;
		if (key == keyset[2]) mark_y--;
		if (key == keyset[3]) mark_y++;
		if (key == 'q') return 1;

		if (key == ' ' && !selected)
		{
			if (grid[mark_x][mark_y]==mark)
			{
				status("Choose destination.");
				sx=mark_x; sy=mark_y; selected=1;
				grid_draw_inverse_mark(sx,sy);
				show_possible_moves(sx,sy);
			} else
			{
				status("Invalid piece. Please choose another.");
				printf(VT_BEEP);
			}
			key='\0';
		}

		if (key == ' ' && selected)
		{
			if (mark_x==sx && mark_y==sy)
			{
				grid_render(mark);
				status("Select piece to move.");
				selected=0;
			}
			else
			if (is_possible_move(mark_x,mark_y, sx,sy))
			{
				process_move(sx,sy, mark_x,mark_y, mark);
				return 0;
			} else
			{
				status("Invalid move. Please retry.");
				printf(VT_BEEP);
			}
			key='\0';
		}

		if (mark_x<0) mark_x=0;
		if (mark_y<0) mark_y=0;

		if (mark_x>GRIDXSIZE-1) mark_x=GRIDXSIZE-1;
		if (mark_y>GRIDYSIZE-1) mark_y=GRIDYSIZE-1;

		grid_draw_inverse_mark(mark_x, mark_y);
		vt_cursor_set((mark_x*2+1), (mark_y+1));
		fflush(stdout);
	}
	return 0;
}

